# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import json
import os
import subprocess
import tempfile
import textwrap

import pytest
from conftest import create_project

# Sample test file content to create for testing
TEST_FILE_CONTENT = """
import pytest

@pytest.mark.parametrize('target', ['esp32', 'linux'], indirect=True)
@pytest.mark.generic
def test_single_target(dut) -> None:
    pass

@pytest.mark.parametrize('target', ['esp32', 'esp32c3'], indirect=True)
@pytest.mark.qemu
def test_single_target_qemu(dut) -> None:
    pass

@pytest.mark.parametrize(
    'count, target',
    [
        (2, 'esp32|esp32s2'),
        (3, 'esp32|esp32|esp32s2'),
        (2, 'linux'),
    ],
)
@pytest.mark.multiboard
def test_multi_dut(dut) -> None:
    pass
"""


# in total should be
# esp32 - generic:
#   test_single_target
# linux - generic:
#   test_single_target
# esp32 - qemu:
#   test_single_target_qemu
# esp32c3 - qemu:
#   test_single_target_qemu
# esp32,esp32s2 - multiboard:
#   test_multi_dut
# esp32,esp32,esp32s2 - multiboard:
#   test_multi_dut
# linux,linux - multiboard:
#   test_multi_dut


@pytest.fixture
def sample_test_file():
    """Create a temporary test file for testing the collector."""
    with tempfile.TemporaryDirectory() as tmpdir:
        # Create pytest.ini with env_markers
        with open(os.path.join(tmpdir, 'pytest.ini'), 'w') as f:
            f.write("""
[pytest]
env_markers =
    generic: applicable to generic ESP devices
    host_test: this test case runs on host machines
""")

        # Create the test file
        test_file_path = os.path.join(tmpdir, 'test_sample.py')
        with open(test_file_path, 'w') as f:
            f.write(TEST_FILE_CONTENT)

        yield test_file_path


class TestCollectFunction:
    @pytest.fixture(autouse=True)
    def setup_test_project(self, tmp_path, test_case_name):
        create_project('sample', tmp_path)
        with open('pytest.ini', 'w') as f:
            f.write("""
[pytest]
env_markers =
    generic: applicable to generic ESP devices
    multiboard: test case runs on multiple ESP devices
""")
        with open(f'{test_case_name}.py', 'w') as f:
            f.write(TEST_FILE_CONTENT)

    def test_output_as_string(self):
        assert (
            textwrap.dedent("""
                esp32,esp32,esp32s2 - multiboard: 1 cases
                \t('esp32', 'esp32', 'esp32s2').('default', 'default', 'default').test_multi_dut
                esp32,esp32s2 - multiboard: 1 cases
                \t('esp32', 'esp32s2').('default', 'default').test_multi_dut
                esp32 - generic: 1 cases
                \tesp32.default.test_single_target
            """).strip()
            == subprocess.check_output(
                ['idf-ci', 'test', 'collect'],
                encoding='utf8',
            ).strip()
        )

    def test_output_as_github_ci(self, test_case_name):
        assert {
            'include': [
                {
                    'targets': 'esp32,esp32,esp32s2',
                    'env_markers': 'multiboard',
                    'runner_tags': ['self-hosted', 'esp32_2+esp32s2', 'multiboard'],
                    'nodes': f'{test_case_name}.py::test_multi_dut[3-esp32|esp32|esp32s2]',
                },
                {
                    'targets': 'esp32,esp32s2',
                    'env_markers': 'multiboard',
                    'runner_tags': ['self-hosted', 'esp32+esp32s2', 'multiboard'],
                    'nodes': f'{test_case_name}.py::test_multi_dut[2-esp32|esp32s2]',
                },
                {
                    'targets': 'linux,linux',
                    'env_markers': 'multiboard',
                    'runner_tags': ['self-hosted', 'linux_2', 'multiboard'],
                    'nodes': f'{test_case_name}.py::test_multi_dut[2-linux]',
                },
                {
                    'targets': 'esp32',
                    'env_markers': 'generic',
                    'runner_tags': ['self-hosted', 'esp32', 'generic'],
                    'nodes': f'{test_case_name}.py::test_single_target[esp32]',
                },
                {
                    'targets': 'esp32',
                    'env_markers': '',
                    'runner_tags': ['self-hosted', 'esp32'],
                    'nodes': f'{test_case_name}.py::test_single_target_qemu[esp32]',
                },
                {
                    'targets': 'esp32c3',
                    'env_markers': '',
                    'runner_tags': ['self-hosted', 'esp32c3'],
                    'nodes': f'{test_case_name}.py::test_single_target_qemu[esp32c3]',
                },
                {
                    'targets': 'linux',
                    'env_markers': 'generic',
                    'runner_tags': ['self-hosted', 'linux', 'generic'],
                    'nodes': f'{test_case_name}.py::test_single_target[linux]',
                },
            ]
        } == json.loads(
            subprocess.check_output(
                ['idf-ci', 'test', 'collect', '--format', 'github', '--marker-expr', ''],
                encoding='utf8',
            ).strip()
        )  # ensure the output is valid JSON



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import typing as t
from collections import defaultdict
from functools import lru_cache
from typing import NamedTuple

from _pytest.python import Function
from pytest_embedded.plugin import parse_multi_dut_args
from pytest_embedded.utils import targets_to_marker

from idf_ci.utils import to_list

logger = logging.getLogger(__name__)


class GroupKey(NamedTuple):
    target_selector: str
    env_selector: str
    runner_tags: t.Tuple[str, ...]


class PytestApp:
    """
    Represents a pytest app.
    """

    def __init__(self, path: str, target: str, config: str) -> None:
        self.path = os.path.abspath(path)
        self.target = target
        self.config = config or 'default'

    def __hash__(self) -> int:
        return hash((self.path, self.target, self.config))

    @property
    def build_dir(self) -> str:
        """
        Returns the build directory for the app.

        .. note::

            Matches the build_dir (by default build_@t_@w) in the idf-build-apps config file.

        :return: The build directory for the app.
        """
        return os.path.join(self.path, f'build_{self.target}_{self.config}')


class PytestCase:
    """
    Represents a pytest test case.
    """

    KNOWN_ENV_MARKERS: t.ClassVar[t.Set[str]] = set()

    def __init__(self, apps: t.List[PytestApp], item: Function) -> None:
        self.apps = apps
        self.item = item

    @classmethod
    def get_param(cls, item: Function, key: str, default: t.Any = None) -> t.Any:
        # funcargs is not calculated while collection
        # callspec is something defined in parametrize
        if not hasattr(item, 'callspec'):
            return default

        return item.callspec.params.get(key, default) or default

    @classmethod
    def from_item(cls, item: Function) -> t.Optional['PytestCase']:
        """
        Turn pytest item to PytestCase
        """
        count = cls.get_param(item, 'count', 1)

        # default app_path is where the test script locates
        app_paths = to_list(parse_multi_dut_args(count, cls.get_param(item, 'app_path', os.path.dirname(item.path))))
        configs = to_list(parse_multi_dut_args(count, cls.get_param(item, 'config', 'default')))
        targets = to_list(parse_multi_dut_args(count, cls.get_param(item, 'target', None)))  # defined fixture

        if targets is None or targets == [None] * count:
            logger.critical(
                'No "target" defined in `@pytest.mark.parametrize` for test function "%s" in file "%s. Skipping...',
                item.name,
                item.path,
            )
            return None

        return PytestCase(
            apps=[PytestApp(app_paths[i], targets[i], configs[i]) for i in range(count)],
            item=item,
        )

    def __hash__(self) -> int:
        return hash((self.path, self.name, self.apps, self.all_markers))

    @property
    def path(self) -> str:
        return str(self.item.path)

    @property
    def name(self) -> str:
        return self.item.originalname

    @property
    def targets(self) -> t.List[str]:
        # sequence matters
        return [app.target for app in self.apps]

    @property
    def target_selector(self) -> str:
        return ','.join(self.targets)

    @property
    def env_markers(self) -> t.Set[str]:
        return {marker for marker in self.all_markers if marker in self.KNOWN_ENV_MARKERS}

    @property
    def env_selector(self) -> str:
        # sequence not matters, sorted for consistency
        return ' and '.join(sorted(self.env_markers))  # pytest marker use 'and' to combine multiple markers

    @property
    def runner_tags(self) -> t.Tuple[str, ...]:
        return tuple(
            [
                targets_to_marker(self.targets),
                *self.env_markers,
            ]
        )

    @property
    def configs(self) -> t.List[str]:
        return [app.config for app in self.apps]

    @property
    def caseid(self) -> str:
        if self.is_single_dut:
            return f'{self.targets[0]}.{self.configs[0]}.{self.name}'
        else:
            return f'{tuple(self.targets)}.{tuple(self.configs)}.{self.name}'

    @property
    def is_single_dut(self) -> bool:
        return True if len(self.apps) == 1 else False

    @property
    def is_host_test(self) -> bool:
        return 'host_test' in self.all_markers or 'linux' in self.targets

    @property
    def is_in_ci(self) -> bool:
        return 'CI_JOB_ID' in os.environ or 'GITHUB_ACTIONS' in os.environ

    @property
    def all_markers(self) -> t.Set[str]:
        return {marker.name for marker in self.item.iter_markers()}

    def get_skip_reason_if_not_built(self, app_dirs: t.Optional[t.List[str]] = None) -> t.Optional[str]:
        """
        Check if all binaries of the test case are built in the app lists.

        :param app_dirs: app folder paths to check
        :return: debug string if not all binaries are built in the app lists, None otherwise
        """
        if app_dirs is None:
            # ignore this feature
            return None

        bin_found = [0] * len(self.apps)
        for i, app in enumerate(self.apps):
            if app.build_dir in app_dirs:
                bin_found[i] = 1

        if sum(bin_found) == 0:
            msg = f'Skip test case {self.name} because all following binaries are not listed in the app lists: '
            for app in self.apps:
                msg += f'\n - {app.build_dir}'

            return msg

        if sum(bin_found) == len(self.apps):
            return None

        # some found, some not, looks suspicious
        msg = f'Found some binaries of test case {self.name} are not listed in the app lists.'
        for i, app in enumerate(self.apps):
            if bin_found[i] == 0:
                msg += f'\n - {app.build_dir}'

        msg += '\nMight be a issue of .build-test-rules.yml files'
        return msg


class GroupedPytestCases:
    def __init__(self, cases: t.List[PytestCase]) -> None:
        self.cases = cases

    @property
    @lru_cache()
    def grouped_cases(self) -> t.Dict[GroupKey, t.List[PytestCase]]:
        """
        Group test cases by target and env_markers
        """
        grouped: t.Dict[GroupKey, t.List[PytestCase]] = defaultdict(list)
        for case in self.cases:
            key = GroupKey(case.target_selector, case.env_selector, case.runner_tags)
            grouped[key].append(case)
        return grouped

    def output_as_string(self) -> str:
        """
        Output as string, for previewing each group of test cases
        """
        lines: t.List[str] = []
        for key, cases in self.grouped_cases.items():
            if not key.env_selector:
                lines.append(f'{key.target_selector}: {len(cases)} cases')
            else:
                lines.append(f'{key.target_selector} - {key.env_selector}: {len(cases)} cases')

            for c in cases:
                lines.append(f'\t{c.caseid}')

        return '\n'.join(lines)

    def output_as_github_ci(self) -> str:
        """
        Output test cases in JSON string for GitHub Actions CI - matrix strategy

        Example output:

        .. code:: json

            {
                include: [
                    {
                        'targets': 'esp32,esp32',
                        'env_markers: 'generic and flash_4mb',
                        'runner_tags': ['self_hosted', 'esp32_2', 'generic', 'flash_4mb'],
                        'nodes': 'nodeid1 nodeid2',
                    },
                    {
                        'targets': 'esp32,esp32s2',
                        'env_markers: 'generic and two_duts',
                        'runner_tags': ['self_hosted', 'esp32+esp32s2', 'generic', 'two_duts'],
                        'nodes': 'nodeid1 nodeid2',
                    },
                    {
                        'targets': 'esp32',
                        'env_markers: 'generic',
                        'runner_tags': ['self_hosted', 'esp32', 'generic'],
                        'nodes': 'nodeid1',
                    ...
                ]
            }

        Example usage:

        .. code:: yaml

            strategy:
                matrix: ${{fromJson( this output as env var )}}
        """
        include_list = []
        for key, cases in self.grouped_cases.items():
            include_list.append(
                {
                    'targets': key.target_selector,
                    'env_markers': key.env_selector,
                    'runner_tags': ['self-hosted', *key.runner_tags],  # self-hosted is required in github
                    'nodes': ' '.join([c.item.nodeid for c in cases]),
                }
            )

        return json.dumps({'include': include_list})

    def output_as_gitlab_ci(self) -> str:
        raise NotImplementedError



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import re
import sys
import typing as t
from pathlib import Path

from idf_build_apps import App, json_to_app
from idf_build_apps.constants import BuildStatus
from pydantic_settings import (
    BaseSettings,
    InitSettingsSource,
    PydanticBaseSettingsSource,
)

from idf_ci._compat import PathLike

logger = logging.getLogger(__name__)


class TomlConfigSettingsSource(InitSettingsSource):
    """
    A source class that loads variables from a TOML file
    """

    def __init__(
        self,
        settings_cls: t.Type[BaseSettings],
        toml_file: t.Optional[PathLike] = Path(''),
    ):
        self.toml_file_path = self._pick_toml_file(
            toml_file,
            '.idf_ci.toml',
        )
        self.toml_data = self._read_file(self.toml_file_path)
        super().__init__(settings_cls, self.toml_data)

    def _read_file(self, path: t.Optional[Path]) -> t.Dict[str, t.Any]:
        if not path or not path.is_file():
            return {}

        if sys.version_info < (3, 11):
            from tomlkit import load

            with open(path) as f:
                return load(f)
        else:
            import tomllib

            with open(path, 'rb') as f:
                return tomllib.load(f)

    @staticmethod
    def _pick_toml_file(provided: t.Optional[PathLike], filename: str) -> t.Optional[Path]:
        """
        Pick a file path to use. If a file path is provided, use it. Otherwise, search up the directory tree for a
        file with the given name.

        :param provided: Explicit path provided when instantiating this class.
        :param filename: Name of the file to search for.
        """
        if provided:
            provided_p = Path(provided)
            if provided_p.is_file():
                fp = provided_p.resolve()
                logger.debug(f'Loading config file: {fp}')
                return fp

        rv = Path.cwd()
        while len(rv.parts) > 1:
            fp = rv / filename
            if fp.is_file():
                logger.debug(f'Loading config file: {fp}')
                return fp

            rv = rv.parent

        return None


class CiSettings(BaseSettings):
    CONFIG_FILE_PATH: t.ClassVar[t.Optional[Path]] = None

    component_mapping_regexes: t.List[str] = [
        '/components/(.+)/',
        '/common_components/(.+)/',
    ]
    extend_component_mapping_regexes: t.List[str] = []

    component_ignored_file_extensions: t.List[str] = [
        '.md',
        '.rst',
        '.yaml',
        '.yml',
        '.py',
    ]
    extend_component_ignored_file_extensions: t.List[str] = []

    # build related settings
    built_app_list_filepatterns: t.List[str] = ['app_info_*.txt']
    preserve_test_related_apps: bool = True
    preserve_non_test_related_apps: bool = False

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: t.Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        dotenv_settings: PydanticBaseSettingsSource,  # noqa: ARG003
        file_secret_settings: PydanticBaseSettingsSource,  # noqa: ARG003
    ) -> t.Tuple[PydanticBaseSettingsSource, ...]:
        sources: t.Tuple[PydanticBaseSettingsSource, ...] = (init_settings,)
        if cls.CONFIG_FILE_PATH is None:
            sources += (TomlConfigSettingsSource(settings_cls, '.idf_ci.toml'),)
        else:
            sources += (TomlConfigSettingsSource(settings_cls, cls.CONFIG_FILE_PATH),)

        return sources

    @property
    def all_component_mapping_regexes(self) -> t.Set[re.Pattern]:
        return {re.compile(regex) for regex in self.component_mapping_regexes + self.extend_component_mapping_regexes}

    def get_modified_components(self, modified_files: t.Iterable[str]) -> t.Set[str]:
        modified_components = set()
        for modified_file in modified_files:
            p = Path(modified_file)
            if p.suffix in self.component_ignored_file_extensions + self.extend_component_ignored_file_extensions:
                continue

            # always use absolute path as posix string
            # Path.resolve return relative path when file does not exist. so use os.path.abspath
            modified_file = Path(os.path.abspath(modified_file)).as_posix()

            for regex in self.all_component_mapping_regexes:
                match = regex.search(modified_file)
                if match:
                    modified_components.add(match.group(1))
                    break

        return modified_components

    def get_apps_list(self) -> t.Optional[t.List[App]]:
        found_files = [p for p in Path('.').glob('app_info_*.txt')]
        if not found_files:
            logger.debug('No built app list files found')
            return None

        logger.debug('Found built app list files: %s', [str(p) for p in found_files])

        apps: t.List[App] = []
        for filepattern in self.built_app_list_filepatterns:
            for filepath in Path('.').glob(filepattern):
                with open(filepath) as fr:
                    for line in fr:
                        line = line.strip()
                        if line:
                            app = json_to_app(line)
                            if app.build_status == BuildStatus.SUCCESS:
                                apps.append(app)
                                logger.debug('App found: %s', apps[-1].build_path)

        if not apps:
            logger.warning(f'No apps found in the built app list files: {found_files}')

        return apps



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import importlib.util
import logging
import os
import re
import sys
import typing as t
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.config import Config
from _pytest.fixtures import FixtureRequest
from _pytest.python import Function
from _pytest.stash import StashKey
from pytest_embedded.plugin import multi_dut_argument, multi_dut_fixture

from ..settings import CiSettings
from ..utils import setup_logging
from .models import PytestCase

_MODULE_NOT_FOUND_REGEX = re.compile(r"No module named '(.+?)'")
IDF_CI_PYTEST_CASE_KEY = StashKey[t.Optional[PytestCase]]()
IDF_CI_PLUGIN_KEY = StashKey['IdfPytestPlugin']()

logger = logging.getLogger(__name__)


############
# Fixtures #
############
@pytest.fixture
@multi_dut_argument
def target(
    request: FixtureRequest,
    target: t.Optional[str],  # noqa: ARG001 # override original target fixture
) -> str:
    _t = getattr(request, 'param', None)
    if not _t:
        raise ValueError('"target" shall either be defined in pytest.mark.parametrize')
    return _t


@pytest.fixture
@multi_dut_argument
def config(request: FixtureRequest) -> str:
    return getattr(request, 'param', None) or 'default'


@pytest.fixture
@multi_dut_fixture
def build_dir(
    request: FixtureRequest,
    app_path: str,
    target: t.Optional[str],
    config: t.Optional[str],
    build_dir: t.Optional[str],  # noqa: ARG001 # override original build_dir fixture
) -> str:
    """
    Check local build dir with the following priority:

    1. build_<target>_<config>
    2. build_<target>
    3. build_<config>
    4. build

    Returns:
        valid build directory
    """
    check_dirs = []
    build_dir_arg = request.config.getoption('build_dir', None)
    if build_dir_arg:
        check_dirs.append(build_dir_arg)
    if target is not None and config is not None:
        check_dirs.append(f'build_{target}_{config}')
    if target is not None:
        check_dirs.append(f'build_{target}')
    if config is not None:
        check_dirs.append(f'build_{config}')
    check_dirs.append('build')

    for check_dir in check_dirs:
        binary_path = os.path.join(app_path, check_dir)
        if os.path.isdir(binary_path):
            logger.info(f'found valid binary path: {binary_path}')
            return check_dir

        logger.warning('checking binary path: %s... missing... try another place', binary_path)

    raise ValueError(
        f'no build dir valid. Please build the binary via "idf.py -B {check_dirs[0]} build" and run pytest again'
    )


def _try_import(path: Path):
    spec = importlib.util.spec_from_file_location('', path)
    # write these if to make mypy happy
    if spec:
        module = importlib.util.module_from_spec(spec)
        if spec.loader and module:
            spec.loader.exec_module(module)


##########
# Plugin #
##########
class IdfPytestPlugin:
    def __init__(
        self,
        *,
        cli_target: str,
        sdkconfig_name: t.Optional[str] = None,
    ) -> None:
        """
        :param cli_target: target passed from command line, could be single target, comma separated targets, or 'all'
        :param sdkconfig_name: run only tests whose apps are built with this sdkconfig name
        """
        self.cli_target = cli_target
        self.sdkconfig_name = sdkconfig_name
        self.apps = CiSettings().get_apps_list()

        self._testing_items: t.Set[pytest.Item] = set()

    @property
    def cases(self) -> t.List[PytestCase]:
        res = []
        for item in self._testing_items:
            c = self.get_case_by_item(item)
            if c:
                res.append(c)

        return sorted(res, key=lambda x: x.caseid)

    @staticmethod
    def get_case_by_item(item: pytest.Item) -> t.Optional[PytestCase]:
        return item.stash.get(IDF_CI_PYTEST_CASE_KEY, None)

    @pytest.hookimpl(tryfirst=True)
    def pytest_pycollect_makemodule(
        self,
        module_path: Path,
    ):
        # no need to install third-party packages for collecting
        # try to eliminate ModuleNotFoundError in test scripts
        while True:
            try:
                _try_import(module_path)
            except ModuleNotFoundError as e:
                res = _MODULE_NOT_FOUND_REGEX.search(e.msg)
                if res:
                    # redirect_stderr somehow breaks the sys.stderr.write() method
                    # fix it when implement proper logging
                    pkg = res.group(1)
                    logger.warning(f'WARNING:Mocking missed package while collecting: {pkg}\n')
                    sys.modules[pkg] = MagicMock()
                    continue
            else:
                break

    @pytest.hookimpl(wrapper=True)
    def pytest_collection_modifyitems(self, config: Config, items: t.List[Function]):
        # add markers definitions
        config.addinivalue_line('markers', 'host_test: this test case runs on host machines')

        for item in items:
            item.stash[IDF_CI_PYTEST_CASE_KEY] = PytestCase.from_item(item)

        # add markers to items
        for item in items:
            _c = self.get_case_by_item(item)
            if _c is None:
                continue

            # add 'host_test' marker to host test cases
            if 'qemu' in _c.all_markers or 'linux' in _c.targets:
                item.add_marker(pytest.mark.host_test)

        yield

        deselected_items: t.List[Function] = []

        # filter by target
        if self.cli_target != 'all':
            res = []
            for item in items:
                _c = self.get_case_by_item(item)
                if _c is None:
                    continue

                if _c.target_selector != self.cli_target:
                    item.add_marker(pytest.mark.skip(reason=f'target mismatch: {self.cli_target}'))
                    deselected_items.append(item)
                else:
                    res.append(item)
            items[:] = res

        # filter by sdkconfig_name
        if self.sdkconfig_name:
            res = []
            for item in items:
                _c = self.get_case_by_item(item)
                if _c is None:
                    continue

                if self.sdkconfig_name not in set(app.config for app in _c.apps):
                    logger.debug('skip test case %s due to sdkconfig name mismatch', _c.caseid)
                    deselected_items.append(item)
                else:
                    res.append(item)
            items[:] = res

        # filter by app list
        if self.apps is not None:
            app_dirs = [os.path.abspath(app.build_path) for app in self.apps]
            res = []
            for item in items:
                _c = self.get_case_by_item(item)
                if _c is None:
                    continue

                skip_reason = _c.get_skip_reason_if_not_built(app_dirs)
                if skip_reason:
                    logger.debug(skip_reason)
                    deselected_items.append(item)
                else:
                    res.append(item)
            items[:] = res

        # deselected items should be added to config.hook.pytest_deselected
        config.hook.pytest_deselected(items=deselected_items)

        self._testing_items.update(items)


##################
# Hook Functions #
##################
def pytest_addoption(parser: pytest.Parser):
    # cli values
    idf_ci_group = parser.getgroup('idf_ci')
    idf_ci_group.addoption(
        '--sdkconfig',
        help='run only tests whose apps are built with this sdkconfig name',
    )

    # ini values
    parser.addini(
        'env_markers',
        help='markers that indicate the running environment of the test case. '
        'Each line is a `<marker_name>: <marker_description>` pair',
        type='linelist',
    )


def pytest_configure(config: Config):
    setup_logging(config.getoption('log_cli_level', None))

    cli_target = config.getoption('target') or 'all'
    sdkconfig_name = config.getoption('sdkconfig', None)

    env_markers: t.Set[str] = set()
    for line in config.getini('env_markers'):
        name, _ = line.split(':', maxsplit=1)
        config.addinivalue_line('markers', line)
        env_markers.add(name)

    PytestCase.KNOWN_ENV_MARKERS = env_markers

    plugin = IdfPytestPlugin(cli_target=cli_target, sdkconfig_name=sdkconfig_name)
    config.stash[IDF_CI_PLUGIN_KEY] = plugin
    config.pluginmanager.register(plugin)


def pytest_unconfigure(config: Config):
    _idf_ci_plugin = config.stash.get(IDF_CI_PLUGIN_KEY, None)
    if _idf_ci_plugin:
        del config.stash[IDF_CI_PLUGIN_KEY]
        config.pluginmanager.unregister(_idf_ci_plugin)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0


__all__ = [
    'IDF_CI_PLUGIN_KEY',
    'IDF_CI_PYTEST_CASE_KEY',
    'GroupedPytestCases',
    'IdfPytestPlugin',
    'PytestCase',
    'get_pytest_cases',
]

from idf_ci.idf_pytest.models import GroupedPytestCases, PytestCase
from idf_ci.idf_pytest.plugin import IDF_CI_PLUGIN_KEY, IDF_CI_PYTEST_CASE_KEY, IdfPytestPlugin
from idf_ci.idf_pytest.scripts import get_pytest_cases



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import typing as t

PathLike = t.Union[str, os.PathLike]


class Undefined(str):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, 'undefined')


UNDEF = Undefined()



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

__all__ = ['click_cli']


import logging
import os

import click

from idf_ci.cli._options import create_config_file
from idf_ci.cli.build_group import build
from idf_ci.cli.test_group import test
from idf_ci.settings import CiSettings
from idf_ci.utils import setup_logging

logger = logging.getLogger(__name__)


_CLI_SETTINGS = {
    'show_default': True,
    'help_option_names': ['-h', '--help'],
}


@click.group(context_settings=_CLI_SETTINGS)
@click.option(
    '-c',
    '--config-file',
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
    help='Path to the idf-ci config file',
)
@click.option(
    '--verbose',
    '-v',
    default=0,
    count=True,
    help='Increase verbosity, can be used multiple times. -v for info, -vv for debug, not set for warning',
)
def click_cli(config_file, verbose):
    if verbose == 0:
        setup_logging(logging.WARNING)
    elif verbose == 1:
        setup_logging(logging.INFO)
    else:
        setup_logging(logging.DEBUG)

    if config_file:
        logger.debug(f'Using config file: {config_file}')
        CiSettings.CONFIG_FILE_PATH = config_file


@click_cli.command()
@click.option('--path', help='Path to create the config file')
def init(path: str):
    """
    Create .idf_ci.toml with default values
    """
    create_config_file(os.path.join(os.path.dirname(__file__), '..', 'templates', '.idf_ci.toml'), path)


@click_cli.command()
def completions():
    """
    Instructions to enable shell completions for idf-ci
    """

    help_message = """
    To enable autocomplete run the following command:

    Bash:
      1. Run this command once

        _IDF_CI_COMPLETE=bash_source idf-ci > ~/.idf-ci-complete.bash

      2. Add the following line to your .bashrc
        . ~/.idf-ci-complete.bash

    Zsh:
      1. Run this command once

        _IDF_CI_COMPLETE=zsh_source idf-ci > ~/.idf-ci-complete.zsh

      2. Add the following line to your .zshrc

        . ~/.idf-ci-complete.zsh

    Fish:
      1. Run this command once

        _IDF_CI_COMPLETE=fish_source idf-ci > ~/.config/fish/completions/idf-ci.fish

    After modifying the shell config, you need to start a new shell in order for the changes to be loaded.
    """
    click.echo(help_message)


click_cli.add_command(build)
click_cli.add_command(test)



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
import textwrap
from pathlib import Path

import pytest
from conftest import create_project
from idf_build_apps.constants import SUPPORTED_TARGETS

from idf_ci import get_all_apps
from idf_ci.cli import click_cli


@pytest.mark.skipif(os.getenv('IDF_PATH') is None, reason='IDF_PATH is set')
class TestGetAllApps:
    @pytest.fixture(autouse=True)
    def _setup(self, runner):
        assert runner.invoke(click_cli, ['build', 'init']).exit_code == 0
        assert runner.invoke(click_cli, ['test', 'init']).exit_code == 0

        yield

    def test_without_test_scripts(self, tmp_path: Path) -> None:
        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)])

        assert len(test_related_apps) == 0
        assert len(non_test_related_apps) == 2 * len(SUPPORTED_TARGETS)
        assert SUPPORTED_TARGETS == [
            'esp32',
            'esp32s2',
            'esp32c3',
            'esp32s3',
            'esp32c2',
            'esp32c6',
            'esp32h2',
            'esp32p4',
        ]

    def test_single_dut_test_script(self, tmp_path: Path) -> None:
        create_project('foo', tmp_path)
        with open(tmp_path / 'foo' / 'test_single_dut_test_script.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
                import pytest

                @pytest.mark.parametrize('target', [
                    'esp32',
                    'esp32s2',
                ], indirect=True)
                def test_foo(dut):
                    pass
                """)
            )
        create_project('bar', tmp_path)

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='all')

        assert len(test_related_apps) == 2
        assert len(non_test_related_apps) == 2 * len(SUPPORTED_TARGETS) - 2

    def test_multi_dut_test_script(self, tmp_path: Path) -> None:
        create_project('foo', tmp_path)
        with open(tmp_path / 'foo' / 'test_multi_dut_test_script.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
                import pytest

                @pytest.mark.parametrize(
                    'count, target', [
                        (2, 'esp32s2|esp32s3'),
                        (3, 'esp32|esp32s3|esp32'),
                    ], indirect=True
                )
                def test_foo(dut):
                    pass
                """)
            )

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='esp32s2,esp32s3')
        assert len(test_related_apps) == 2
        assert len(non_test_related_apps) == 0

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='esp32,esp32s3,esp32')
        assert len(test_related_apps) == 2
        assert len(non_test_related_apps) == 0

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='all')
        assert len(test_related_apps) == 3
        assert len(non_test_related_apps) == len(SUPPORTED_TARGETS) - 3

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='foo,bar')
        assert len(test_related_apps) == 0
        assert len(non_test_related_apps) == 0

    def test_modified_pytest_script(self, tmp_path: Path) -> None:
        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        (tmp_path / 'test_modified_pytest_script.py').write_text(
            textwrap.dedent("""
            import pytest
            import os

            @pytest.mark.parametrize('count, target', [(2, 'esp32')], indirect=True)
            @pytest.mark.parametrize('app_path', [
                    '{}|{}'.format(os.path.join(os.path.dirname(__file__), 'foo'),
                                   os.path.join(os.path.dirname(__file__), 'bar')),
                ], indirect=True
            )
            def test_multi_foo_bar(dut):
                pass
            """),
            encoding='utf-8',
        )

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='all')
        assert len(test_related_apps) == 2  # foo-esp32, bar-esp32
        assert len(non_test_related_apps) == 2 * len(SUPPORTED_TARGETS) - 2

        test_related_apps, non_test_related_apps = get_all_apps(
            [str(tmp_path)], target='all', modified_files=[], modified_components=[]
        )
        assert len(test_related_apps) == 0
        assert len(non_test_related_apps) == 0

        test_related_apps, non_test_related_apps = get_all_apps(
            [str(tmp_path)],
            target='all',
            modified_files=[str(tmp_path / 'test_modified_pytest_script.py')],
            modified_components=[],
        )
        assert len(test_related_apps) == 2
        assert len(non_test_related_apps) == 0

    def test_host_test_script(self, tmp_path: Path) -> None:
        create_project('foo', tmp_path)
        (tmp_path / 'foo' / 'sdkconfig.ci').touch()
        (tmp_path / 'foo' / 'sdkconfig.ci.linux').write_text('CONFIG_IDF_TARGET="linux"\n', encoding='utf-8')

        with open(tmp_path / 'foo' / 'test_host_test_script.py', 'w') as fw:
            fw.write(
                textwrap.dedent("""
                import pytest

                @pytest.mark.parametrize('target', [
                    'linux',
                ], indirect=True)
                def test_foo(dut):
                    pass

                @pytest.mark.parametrize('target', [
                    'esp32',
                ], indirect=True)
                @pytest.mark.qemu
                def test_foo_qemu(dut):
                    pass
                """)
            )

        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='all')

        assert len(test_related_apps) == 0
        assert len(non_test_related_apps) == len(SUPPORTED_TARGETS)

        # by default, linux is not built
        test_related_apps, non_test_related_apps = get_all_apps([str(tmp_path)], target='all', marker_expr='host_test')
        assert len(test_related_apps) == 1
        assert len(non_test_related_apps) == len(SUPPORTED_TARGETS) - 1

        # specify linux
        test_related_apps, non_test_related_apps = get_all_apps(
            [str(tmp_path)],
            target='linux',
        )
        assert len(test_related_apps) == 1
        assert len(non_test_related_apps) == 0



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import fnmatch
import logging
import os
import typing as t

from idf_build_apps import App, build_apps, find_apps
from idf_build_apps.constants import SUPPORTED_TARGETS, BuildStatus

from . import get_pytest_cases
from ._compat import UNDEF, Undefined
from .settings import CiSettings

logger = logging.getLogger(__name__)


def get_all_apps(
    paths: t.List[str],
    target: str = 'all',
    *,
    modified_files: t.Optional[t.List[str]] = None,
    modified_components: t.Optional[t.List[str]] = None,
    marker_expr: str = UNDEF,
    default_build_targets: t.List[str] = UNDEF,  # type: ignore
) -> t.Tuple[t.Set[App], t.Set[App]]:
    apps = []
    for _t in target.split(','):
        if isinstance(default_build_targets, Undefined):
            if _t != 'all' and _t not in SUPPORTED_TARGETS:
                default_targets = list({*SUPPORTED_TARGETS, _t})
            else:
                default_targets = SUPPORTED_TARGETS
        else:
            default_targets = default_build_targets

        apps.extend(
            find_apps(
                paths,
                _t,
                modified_files=modified_files,
                modified_components=modified_components,
                include_skipped_apps=True,
                default_build_targets=default_targets,
            )
        )

    cases = get_pytest_cases(paths, target, marker_expr=marker_expr)
    if not cases:
        return set(), set(apps)

    # Get modified pytest cases if any
    modified_pytest_cases = []
    if modified_files:
        modified_pytest_scripts = [
            os.path.dirname(f) for f in modified_files if fnmatch.fnmatch(os.path.basename(f), 'test_*.py')
        ]
        if modified_pytest_scripts:
            modified_pytest_cases = get_pytest_cases(modified_pytest_scripts, target, marker_expr=marker_expr)

    # Create dictionaries mapping app info to test cases
    def get_app_dict(cases):
        return {(case_app.path, case_app.target, case_app.config): case for case in cases for case_app in case.apps}

    pytest_dict = get_app_dict(cases)
    modified_pytest_dict = get_app_dict(modified_pytest_cases)

    test_related_apps = set()
    non_test_related_apps = set()

    for app in apps:
        app_key = (os.path.abspath(app.app_dir), app.target, app.config_name or 'default')
        # override build_status if test script got modified\
        _case = modified_pytest_dict.get(app_key)
        if _case:
            test_related_apps.add(app)
            app.build_status = BuildStatus.SHOULD_BE_BUILT
            logger.debug('Found app: %s - required by modified test case %s', app, _case.path)
        elif app.build_status != BuildStatus.SKIPPED:
            _case = pytest_dict.get(app_key)
            if _case:
                test_related_apps.add(app)
                # build or not should be decided by the build stage
                logger.debug('Found test-related app: %s - required by %s', app, _case.path)
            else:
                non_test_related_apps.add(app)
                logger.debug('Found non-test-related app: %s', app)

    return test_related_apps, non_test_related_apps


def build(
    paths: t.List[str],
    target: str,
    *,
    parallel_count: int = 1,
    parallel_index: int = 1,
    modified_files: t.Optional[t.List[str]] = None,
    only_test_related: bool = False,
    only_non_test_related: bool = False,
    dry_run: bool = False,
    verbose: t.Optional[int] = None,
) -> t.Tuple[t.List[App], int]:
    modified_components = None
    if modified_files is not None:
        modified_components = sorted(CiSettings().get_modified_components(modified_files))
        logger.debug('Modified files: %s', modified_files)
        logger.debug('Modified components: %s', modified_components)

    # we have to call get_all_apps first, then call build_apps(apps)
    test_related_apps, non_test_related_apps = get_all_apps(
        paths,
        target,
        modified_files=modified_files,
        modified_components=modified_components,
    )

    for _app in test_related_apps:
        _app.preserve = CiSettings().preserve_test_related_apps

    for _app in non_test_related_apps:
        _app.preserve = CiSettings().preserve_non_test_related_apps

    if not only_test_related and not only_non_test_related:
        apps = sorted(test_related_apps.union(non_test_related_apps))
    elif only_test_related:
        apps = sorted(test_related_apps)
    else:
        apps = sorted(non_test_related_apps)

    ret = build_apps(
        apps,
        parallel_count=parallel_count,
        parallel_index=parallel_index,
        dry_run=dry_run,
        modified_files=modified_files,
        modified_components=modified_components,
        verbose=verbose,
    )
    return apps, ret



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import logging
import os

import click

from idf_ci import get_pytest_cases

from ..idf_pytest.models import GroupedPytestCases
from ._options import create_config_file

logger = logging.getLogger(__name__)


@click.group()
def test():
    """
    Group of test related commands
    """
    pass


@test.command()
@click.option('--path', help='Path to create the config file')
def init(path: str):
    """
    Create pytest.ini with default values
    """
    create_config_file(os.path.join(os.path.dirname(__file__), '..', 'templates', 'pytest.ini'), path)


@test.command()
@click.argument('paths', nargs=-1, type=click.Path(dir_okay=False, file_okay=True, exists=True))
@click.option('-m', '--marker-expr', default='not host_test', help='Pytest marker expression')
@click.option('-t', '--target', default='all', help='Target to be processed. Or "all" to process all targets.')
@click.option(
    '--format',
    '_format',
    type=click.Choice(
        [
            'raw',
            'github',
            # 'gitlab',  # TODO
        ]
    ),
    default='raw',
    help='Output format',
)
@click.option(
    '-o',
    '--output',
    type=click.Path(dir_okay=False, file_okay=True),
    help='Output destination. Stdout if not provided',
)
def collect(paths, *, marker_expr, target, _format, output):
    """
    Run pytest with provided arguments
    """
    logger.debug(f'Collecting test cases from {paths} with target {target} and marker expression {marker_expr}')

    cases = get_pytest_cases(
        paths=paths or ['.'],
        target=target or 'all',
        marker_expr=marker_expr,
    )

    grouped_cases = GroupedPytestCases(cases)

    if _format == 'raw':
        s = grouped_cases.output_as_string()
    elif _format == 'github':
        s = grouped_cases.output_as_github_ci()
    elif _format == 'gitlab':
        s = grouped_cases.output_as_gitlab_ci()
    else:
        raise ValueError(f'Unknown output format: {_format}')

    if output is None:
        click.echo(s)
    else:
        with open(output, 'w') as f:
            f.write(s)

        click.echo(f'Created test cases collection file: {output}')



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import io
import logging
import os.path
import typing as t
from contextlib import redirect_stderr, redirect_stdout

import pytest
from _pytest.config import ExitCode

from idf_ci._compat import UNDEF, Undefined

from ..utils import remove_subfolders, setup_logging
from .models import PytestCase
from .plugin import IdfPytestPlugin

logger = logging.getLogger(__name__)


def get_pytest_cases(
    paths: t.List[str],
    target: str = 'all',
    *,
    sdkconfig_name: t.Optional[str] = None,
    marker_expr: str = UNDEF,
) -> t.List[PytestCase]:
    if isinstance(marker_expr, Undefined):
        if 'linux' in target:
            marker_expr = 'host_test'
        else:
            marker_expr = 'not host_test'

    plugin = IdfPytestPlugin(
        cli_target=target,
        sdkconfig_name=sdkconfig_name,
    )
    args = [
        # remove sub folders if parent folder is already in the list
        # https://github.com/pytest-dev/pytest/issues/13319
        *remove_subfolders(paths),
        '--collect-only',
        '--rootdir',
        os.getcwd(),
        '--target',
        target,
    ]
    if marker_expr:
        args.extend(['-m', f'{marker_expr}'])

    cur_log_level = logger.parent.level  # type: ignore

    with io.StringIO() as out_b, io.StringIO() as err_b:
        with redirect_stdout(out_b), redirect_stderr(err_b):
            res = pytest.main(args, plugins=[plugin])
        stdout_msg = out_b.getvalue()
        stderr_msg = err_b.getvalue()

    setup_logging(level=cur_log_level)  # the redirect messes up the logging level

    if res == ExitCode.OK:
        return plugin.cases

    raise RuntimeError(f'pytest collection failed.\nArgs: {args}\nStdout: {stdout_msg}\nStderr: {stderr_msg}')
